/* mdyndns.c - a minimal dyndns.org client
 *
 * (C) 2004 Thomas Knop <tknop@maxrelax.de>
 *
 * $Id: mdyndns.c,v 1.4 2005/12/03 22:07:29 tknop Exp $
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdlib.h>
#include <unistd.h>
#include <limits.h>
#include <time.h>
#include <string.h>
#include <buffer.h>
#include <fmt.h>
#include <scan.h>
#include <netdb.h>
#include <textcode.h>
#include <netinet/in.h>

#include "config.h"
#ifdef USE_SYSLOG
#include <sys/syslog.h>
#endif

#define CLIENT  "mdyndns"
#define VERSION "0.2.1"
#define TARGET  "members.dyndns.org"

char *user;
char *passwd;
char *host;

char *sys      = "dyndns";
char *myip     = 0;
char *mx       = 0;
char *wildcard = "off";
char *backmx   = 0;
char *offline  = 0;

char *prog;

#define buffer_puts2(b,s1,s2) buffer_puts3((b),(s1),(s2),0)

static void buffer_puts3(buffer *b, char *s1, char *s2, char *s3)
{
  if (s1) buffer_puts(b,s1);
  if (s2) buffer_puts(b,s2);
  if (s3) buffer_puts(b,s3);
}

static void log2(int err, char *s1, char *s2)
{
#ifdef USE_SYSLOG
  syslog(err ? LOG_ERR : LOG_NOTICE, "%s %s", 
         s1 ? s1 : "", s2 ? s2 : "");
  if (err) {
    closelog();
    exit(1);
  }
#else
  buffer *b = err ? buffer_2 : buffer_1;
  buffer_puts3(b, prog, ": ", s1);
  buffer_puts2(b, s2, "\n");
  buffer_flush(b);
  if (err) exit(1);
#endif
}

void usage()
{
  buffer_puts3(buffer_2, 
	       "usage: ", prog, " [-wh] [-s system] [-i ip] [-m mx] [-b backmx] [-o offline] user passwd hostname\n");
  buffer_flush(buffer_2);
  exit(1);
}

static int open_socked(char *host, int port)
{
  char buf[4000];
  struct hostent h_ent, *h;
  int ht_err;
  struct sockaddr_in sa;
  int sockfd;

  memset(&sa, 0, sizeof sa);
  if (gethostbyname_r(host, &h_ent, buf, sizeof buf, &h, &ht_err))
    log2(1, "can't resolve host ", host);
  sa.sin_family = h->h_addrtype;
  memmove(&sa.sin_addr, h->h_addr, h->h_length);
  sa.sin_port = htons(port);
  sockfd = socket(sa.sin_family, SOCK_STREAM, 0);
  if (sockfd < 0)
    log2(1, "can't create socket", 0);
  if (connect(sockfd, (struct sockaddr*) &sa, sizeof sa) < 0)
    log2(1, "can't connect to host ", host);
  return sockfd;
}

#ifdef PROXY_SUPPORT
int proxy_connect()
{
  char *e, *host;
  long port;

  if ((e = getenv("http_proxy")) != NULL) {
    if (strncmp(e, "http://", 7))
      log2(1, "invalid http_proxy environment", 0);
    e += 7;
    host = e;
    port = 8080;
    while(*e) {
      if (*e == ':') {
	scan_long(e+1, &port);
	break;
      }
      if (*e == '/')
	break;
      ++e;
    }
    *e = 0;
  } else {
    host = TARGET;
    port = 80;
  }
  return open_socked(host, (int)port);
}
#endif

int stradd(char *dest, const char *src)
{
  int len = strlen(src);
  memcpy(dest, src, len);
  return len;
}

void doit(char *user, char *passwd, char *host)
{
  char buf[4000];
  char buf2[100];
  char *t, *u;
#ifdef PROXY_SUPPORT
  int fd = proxy_connect();
#else
  int fd = open_socked(TARGET, 80);
#endif

  t = buf;
  /* request line */
  t += stradd(t, "GET ");
#ifdef PROXY_SUPPORT
  t += stradd(t, "http://");
  t += stradd(t, TARGET);
  t += stradd(t, ":80");
#endif
  t += stradd(t, "/nic/update?system=");
  t += stradd(t, sys);
  t += stradd(t, "&hostname=");
  t += stradd(t, host);
  t += stradd(t, "&wildcard=");
  t += stradd(t, wildcard);
  if (myip) { t += stradd(t, "&myip="); t += stradd(t, myip); }
  if (mx) { t += stradd(t, "&mx="); t += stradd(t, mx); }
  if (backmx) { t += stradd(t, "&backmx="); t += stradd(t, backmx); }
  if (offline) { t += stradd(t, "&offline="); t += stradd(t, offline); }
  t += stradd(t, " HTTP/1.0\n");
  /* headers */
  t += stradd(t, "Host: ");
  t += stradd(t, TARGET);
  t += stradd(t, "\n");
  t += stradd(t, "User-Agent: " CLIENT "/" VERSION "\n");
  t += stradd(t, "Authorization: Basic ");
  /* authorization */
  u = buf2;
  u += stradd(u, user);
  u += stradd(u, ":");
  u += stradd(u, passwd);
  t += fmt_base64(t, buf2, u-buf2);
  t += stradd(t, "\n\n");
  /* output the stuff */
#ifdef DEBUG
  buffer_puts(buffer_2, "sending:\n");
  *t = 0;
  buffer_putsflush(buffer_2, buf);
#endif
  write(fd, buf, strlen(buf));
  buf[read(fd, buf, sizeof(buf)-1)] = 0;
#ifdef DEBUG
  buffer_puts(buffer_2, "server answer:\n");
  buffer_putsflush(buffer_2, buf);
#endif
  if (strlen(buf) > 2) {
    t = buf;
    do {
      u = t;
      t++;
      if ((*u == '\n') && (*t == '\r')) break;
    } while (*t);
    if (t) {
#ifdef USE_SYSLOG
      syslog(LOG_NOTICE, "%s", t+2);
#else
      buffer_puts2(buffer_1, prog, ": ");
      buffer_puts(buffer_1, t+2);
      buffer_putsflush(buffer_1, "\n");
#endif
    } else {
      log2(1, "unknown and/or invalid return from", TARGET);
    }
  }
}

int main(int argc, char *argv[])
{
  int c;
  register char *s;

  for (s=prog=argv[0]; *s != '\0';)
    if (*s++ == '/') prog = s;
  while ((c = getopt(argc, argv, "s:i:wm:b:o:")) != -1) {
    switch (c) {
    case 's': sys = optarg; break;
    case 'i': myip = optarg; break;
    case 'w': wildcard = "on"; break;
    case 'm': mx = optarg; break;
    case 'b': backmx = optarg; break;
    case 'o': offline = optarg; break;
    case 'h':
    default:
      usage();
      break;
    }
  }
  if (argc-optind != 3) usage();
#ifdef USE_SYSLOG
  openlog(prog, 0, LOG_USER);
#endif
  doit(argv[optind], argv[optind+1], argv[optind+2]);
  return 0;
}
